from pathlib import Path
from typing import Union

from gspread import service_account, service_account_from_dict


class Connection:
    def __init__(
        self, credentials_path: Union[Path, str] = None, credentials_dict: dict = None
    ):
        self.open = False
        func = service_account if credentials_path else service_account_from_dict
        self.client = func(credentials_path or credentials_dict)



from typing import TYPE_CHECKING, Any

from gspread.utils import ValueInputOption, ValueRenderOption
from gspread_formatting import format_cell_range

from .style import Style
from .utils import run_in_executor

if TYPE_CHECKING:
    from .row import Row
    from .tab import Tab


input_formats = {
    "raw": ValueInputOption.raw,
    "user_entered": ValueInputOption.user_entered,
}

render_formats = {
    "formatted": ValueRenderOption.formatted,
    "unformatted": ValueRenderOption.unformatted,
    "formula": ValueRenderOption.formula,
}


class Cell(str):
    def __new__(
        cls,
        value,
        tab: "Tab",
        label: str,
        row_index: int,
        cell_index: int = 0,
        row: "Row" = None,
    ):
        instance = super().__new__(cls, value)
        instance.tab = tab
        instance.label = label
        instance.row_index = row_index
        instance.cell_index = cell_index
        instance.row = row
        return instance

    async def clear(self):
        if self.row:
            self.row[self.cell_index] = Cell(
                "", self.tab, label=self.label, row_index=self.row_index, row=self.row
            )

        print(self.label, self.row_index)
        await run_in_executor(self.tab.batch_clear, [f"{self.label}{self.row_index}"])

    async def update(
        self, value: Any, input_format: str = "raw", render_format: str = "formatted"
    ):
        if self.row:
            self.row[self.cell_index] = Cell(
                value,
                self.tab,
                label=self.label,
                row_index=self.row_index,
                row=self.row,
            )

        await run_in_executor(
            self.tab.update,
            f"{self.label}{self.row_index}",
            value,
            value_input_option=input_formats.get(input_format, "raw"),
            response_value_render_option=render_formats.get(render_format, "formatted"),
        )
        return Cell(
            value, self.tab, self.label, self.row_index, self.cell_index, self.row
        )

    async def style(self, obj: Style):
        style = obj.raw if isinstance(obj, Style) else obj
        await run_in_executor(
            format_cell_range, self.tab, f"{self.label}{self.row_index}", style
        )

    async def delete(self, shift: str = "left"):
        await self.tab.del_cell(f"{self.label}{self.row_index}", shift=shift)



import asyncio
import math
import string
from functools import partial


def chat_at(length):
    return string.ascii_uppercase[length - 1]


def get_location(length, index=1):
    if length <= 26:
        return f"{chat_at(length)}{index}"

    iteration = math.ceil(length / 26)
    first_char = chat_at(iteration - 1)
    second_char = chat_at(length % 26)

    return f"{first_char}{second_char}"


async def run_in_executor(func, *args, **kwargs):
    func = partial(func, *args, **kwargs)
    data = await asyncio.get_event_loop().run_in_executor(None, func)
    return data



from typing import TYPE_CHECKING, List, Union

from gspread_formatting import CellFormat, format_cell_range

from .cell import Cell
from .style import Style
from .utils import get_location, run_in_executor

if TYPE_CHECKING:
    from .tab import Tab


class Row(list):
    def __init__(self, items, tab: "Tab", index: int):
        self.tab = tab
        self.row_index = index
        self.items = self.convert_to_cell(items)
        super().__init__(self.items)

    def convert_to_cell(self, items):
        def to_cell(item):
            i, value = item
            return Cell(value, self.tab, get_location(i + 1), self.row_index, i, self)

        converted: List[Cell] = list(map(to_cell, enumerate(items)))
        return converted

    async def clear(self) -> None:
        await run_in_executor(
            self.tab.batch_clear, [f"{self.row_index}:{self.row_index}"]
        )
        empty_items = self.convert_to_cell(list(map(lambda _: "", self.items)))
        self.items = empty_items
        super().__init__(empty_items)

    async def update(self, values: list, start: str = "A", *args, **kwargs):
        await run_in_executor(
            self.tab.update, f"{start}{self.row_index}", [values], *args, **kwargs
        )
        self.items = self.convert_to_cell(values)
        super().__init__(self.items)

    async def style(self, obj: Union[Style, CellFormat]):
        style = obj.raw if isinstance(obj, Style) else obj
        await run_in_executor(
            format_cell_range, self.tab, f"{self.row_index}:{self.row_index}", style
        )

    async def refetch(self):
        values = await run_in_executor(self.tab.values)
        row = values[self.row_index - 1]
        self.items = self.convert_to_cell(row)
        super().__init__(self.items)

    async def append_cell(self, value, *args, **kwargs):
        new_values = value if isinstance(value, list) else [value]
        new_data = [*self.items, *new_values]
        await self.update(new_data, *args, **kwargs)

    async def delete(self):
        await self.tab.del_row(self.row_index)



from dataclasses import dataclass
from typing import List

from gspread import Worksheet
from gspread.utils import ValueRenderOption

from .cell import Cell
from .row import Row
from .utils import run_in_executor

render_formats = {
    "formatted": ValueRenderOption.formatted,
    "unformatted": ValueRenderOption.unformatted,
    "formula": ValueRenderOption.formula,
}


@dataclass
class Tab(Worksheet):
    def __init__(self, sheet, *args, **kwargs):
        self.sheet = sheet
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"<Tab title='{self.title}' id={self.id}>"

    async def values(self, **kwargs) -> List[Row]:
        rows = await run_in_executor(self.get_values, **kwargs)

        for i, row in enumerate(rows):
            rows[i] = Row(row, tab=self, index=i + 1)

        return rows

    async def get_row(self, serial_no: int):
        values = await self.values()
        return values[serial_no - 1]

    async def get_cell(
        self, cell_name: str, render_option: str = "formatted", **kwargs
    ):
        resp = await run_in_executor(
            self.get,
            cell_name,
            value_render_option=render_formats.get(render_option, "formatted"),
            **kwargs,
        )

        data = resp[0][0]
        return Cell(
            data,
            tab=self,
            label=cell_name[0],
            row_index=int(cell_name[1:]),
            cell_index=0,
            row=None,
        )

    async def append(self, data: list, get_row: bool = False):
        await run_in_executor(self.append_row, data)

        if get_row:
            values = await self.values()
            return values[-1]

    async def del_row(self, start: int, end: int = None):
        await run_in_executor(
            self.delete_rows, start_index=start, end_index=end or start
        )

    async def del_cell(self, start: str, end: str = None, shift: str = "up"):
        [start_col_name, *start_row_indexs] = list(start)
        if end:
            [end_col_name, *end_row_indexs] = list(end)
        else:
            end_col_name = start_col_name
            end_row_indexs = start_row_indexs

        start_row_index = int("".join(start_row_indexs)) - 1
        end_row_index = int("".join(end_row_indexs))

        start_col_index = ord(start_col_name) - 65
        end_col_index = ord(end_col_name) - 64

        sheet_id = self._properties["sheetId"]
        requests = [
            {
                "deleteRange": {
                    "range": {
                        "sheetId": sheet_id,
                        "startRowIndex": start_row_index,
                        "endRowIndex": end_row_index,
                        "startColumnIndex": start_col_index,
                        "endColumnIndex": end_col_index,
                    },
                    "shiftDimension": "COLUMNS" if shift == "left" else "ROWS",
                }
            }
        ]

        self.sheet.batch_update({"requests": requests})



from dataclasses import dataclass

from gspread import Client, Spreadsheet

from .connection import Connection
from .tab import Tab
from .utils import run_in_executor


@dataclass
class Sheet(Spreadsheet):
    sheet_name: str
    folder_id: str = None
    connection: Connection = None
    sheet: Spreadsheet = None
    _properties: dict = None
    _open: bool = False

    async def open(self):
        if not self._open:
            sheet = await run_in_executor(
                self.connection.client.open, self.sheet_name, folder_id=self.folder_id
            )
            self.sheet = sheet
            self.client: Client = sheet.client
            self._properties: dict = sheet._properties  # noqa
            self._open = True

    async def get_tab(self, tab_name: str) -> Tab:
        await self.open()
        _tab = await run_in_executor(self.worksheet, tab_name)
        return Tab(
            spreadsheet=self.sheet,
            properties=_tab._properties,  # noqa
            sheet=self,
            spreadsheet_id=self.sheet.id,
            client=self.sheet.client,
        )

    async def tabs(self, exclude_hidden: bool = False):
        await self.open()
        sheets = await run_in_executor(self.worksheets, exclude_hidden=exclude_hidden)

        return [
            Tab(
                spreadsheet=self,
                properties=tab._properties,  # noqa
                sheet=self,
                spreadsheet_id=tab.id,
                client=self.sheet.client,
            )
            for tab in sheets
        ]
